# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import datetime
import functools
import inspect
import json
import traceback as tb
from abc import ABCMeta
from enum import Enum
from typing import Any
from typing import Callable

import kolena
from kolena._api.v1.client_log import ClientLog as API
from kolena._api.v1.core import TestRun as CoreAPI
from kolena._utils import krequests
from kolena._utils.state import _client_state


class DatadogLogLevels(str, Enum):
    ERROR = "error"


def telemetry(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not _client_state.telemetry:  # no extra exception handling unless enabled
            return func(*args, **kwargs)
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            log_telemetry(e)
            raise e

    return wrapper


def upload_log(message: str, status: str) -> None:
    request = API.UploadLogRequest(
        client_version=kolena.__version__,
        timestamp=str(datetime.datetime.now()),
        message=message,
        status=status,
    )
    krequests.post(endpoint_path=API.Path.UPLOAD.value, json=dataclasses.asdict(request))


def log_telemetry(e: BaseException) -> None:
    try:
        stack = tb.format_stack()
        exc_format = tb.format_exception(None, e, e.__traceback__)
        combined = stack + exc_format
        upload_log("".join(combined), DatadogLogLevels.ERROR.value)
    except BaseException:
        """
        Attempting to upload the telemetry is best-effort. We don't want to have exceptions in that
        process be thrown to the customer--instead they should get their original stacktrace.
        """
        ...


class WithTelemetry(metaclass=ABCMeta):
    """
    Applies ``@telemetry`` to each "public" (non-underscored) service method declared on a subclass.
    """

    def __init_subclass__(cls) -> None:
        for key, value in cls.__dict__.items():
            if key.startswith("_") and not key == "__init__" or not callable(value) or inspect.isclass(value):
                continue
            setattr(cls, key, telemetry(value))


def report_crash(id: int, endpoint_path: str):
    request = CoreAPI.MarkCrashedRequest(test_run_id=id)
    # note no krequests.raise_for_status -- already in crashed state
    krequests.post(endpoint_path=endpoint_path, data=json.dumps(dataclasses.asdict(request)))



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import math
from typing import Tuple

import numpy as np
import pandas as pd
from PIL import Image
from requests_toolbelt import MultipartEncoder

from kolena._api.v1.fr import Asset as AssetAPI
from kolena._utils import krequests
from kolena._utils.asset_path_mapper import AssetPathMapper
from kolena._utils.consts import BatchSize
from kolena.fr.datatypes import _ImageChipsDataFrame


def upload_image_chips(df: _ImageChipsDataFrame, batch_size: int = BatchSize.UPLOAD_CHIPS.value) -> None:
    def upload_batch(df_batch: _ImageChipsDataFrame) -> None:
        df_batch = df_batch.reset_index(drop=True)  # reset indices so we match the signed_url indices

        def as_path_stub_and_buffer(row: pd.Series) -> Tuple[str, io.BytesIO]:
            pil_image = Image.fromarray(row["image"]).convert("RGB")
            image_buf = io.BytesIO()
            pil_image.save(image_buf, "png")
            image_buf.seek(0)
            return AssetPathMapper.path_stub(row["test_run_id"], row["uuid"], row["image_id"], row["key"]), image_buf

        data = MultipartEncoder(fields=[("files", as_path_stub_and_buffer(row)) for _, row in df_batch.iterrows()])
        upload_response = krequests.put(
            endpoint_path=AssetAPI.Path.BULK_UPLOAD.value,
            data=data,
            headers={"Content-Type": data.content_type},
        )
        krequests.raise_for_status(upload_response)

    num_chunks = math.ceil(len(df) / batch_size)
    chunk_iter = np.array_split(df, num_chunks) if len(df) > 0 else []
    for df_chunk in chunk_iter:
        upload_batch(df_chunk)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from kolena._experimental.object_detection import GroundTruth
from kolena._experimental.object_detection import Inference
from kolena._experimental.object_detection import TestCase
from kolena._experimental.object_detection import TestCaseMetrics
from kolena._experimental.object_detection import TestCaseMetricsSingleClass
from kolena._experimental.object_detection import TestSample
from kolena._experimental.object_detection import TestSampleMetrics
from kolena._experimental.object_detection import TestSampleMetricsSingleClass
from kolena._experimental.object_detection import TestSuite
from kolena._experimental.object_detection import TestSuiteMetrics
from kolena._experimental.object_detection import ThresholdConfiguration
from kolena._experimental.object_detection.evaluator_multiclass import MulticlassObjectDetectionEvaluator
from kolena._experimental.object_detection.evaluator_single_class import SingleClassObjectDetectionEvaluator
from kolena.workflow import Evaluator
from kolena.workflow import Plot


class ObjectDetectionEvaluator(Evaluator):
    """
    This `ObjectDetectionEvaluator` transforms inferences into metrics for the object detection workflow for a
    single class or multiple classes.

    When a [`ThresholdConfiguration`][kolena._experimental.object_detection.workflow.ThresholdConfiguration] is
    configured to use an F1-Optimal threshold strategy, the evaluator requires that the first test case retrieved for
    a test suite contains the complete sample set.

    For additional functionality, see the associated [base class documentation][kolena.workflow.evaluator.Evaluator].
    """

    # The evaluator class to use for single or multiclass object detection
    evaluator: Optional[
        Union[
            SingleClassObjectDetectionEvaluator,
            MulticlassObjectDetectionEvaluator,
        ]
    ] = None

    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> List[Tuple[TestSample, Union[TestSampleMetrics, TestSampleMetricsSingleClass]]]:
        assert configuration is not None, "must specify configuration"

        # Use complete test case to determine workflow, single class or multiclass
        if self.evaluator is None:
            if configuration.with_class_level_metrics:
                self.evaluator = MulticlassObjectDetectionEvaluator()
            else:
                self.evaluator = SingleClassObjectDetectionEvaluator()

        return self.evaluator.compute_test_sample_metrics(
            test_case=test_case,
            inferences=inferences,
            configuration=configuration,
        )

    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[Union[TestSampleMetrics, TestSampleMetricsSingleClass]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> Union[TestCaseMetrics, TestCaseMetricsSingleClass]:
        assert configuration is not None, "must specify configuration"
        return self.evaluator.compute_test_case_metrics(
            test_case=test_case,
            inferences=inferences,
            metrics=metrics,
            configuration=configuration,
        )

    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[Union[TestSampleMetrics, TestSampleMetricsSingleClass]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> Optional[List[Plot]]:
        assert configuration is not None, "must specify configuration"
        return self.evaluator.compute_test_case_plots(
            test_case=test_case,
            inferences=inferences,
            metrics=metrics,
            configuration=configuration,
        )

    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, Union[TestCaseMetrics, TestCaseMetricsSingleClass]]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> TestSuiteMetrics:
        assert configuration is not None, "must specify configuration"
        return self.evaluator.compute_test_suite_metrics(
            test_suite=test_suite,
            metrics=metrics,
            configuration=configuration,
        )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

from pydantic import validate_arguments

from kolena._utils.frozen import Frozen
from kolena._utils.geometry import validate_polygon
from kolena._utils.inference_validators import validate_label
from kolena._utils.serializable import Serializable
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal.ground_truth import GroundTruthType


class GroundTruth(Serializable, Frozen, metaclass=ABCMeta):
    """
    Base class for ground truths associated with an image.

    See concrete implementations [`BoundingBox`][kolena.detection.ground_truth.BoundingBox] and
    [`SegmentationMask`][kolena.detection.ground_truth.SegmentationMask] for details.
    """

    label: str
    difficult: bool

    @abstractmethod
    def __init__(self) -> None:
        ...

    @classmethod
    def _from_dict(cls, ground_truth: Dict[str, Any]) -> "GroundTruth":
        data_type = ground_truth["data_type"]
        data_object = ground_truth["data_object"]
        if data_type == GroundTruthType.CLASSIFICATION_LABEL.value:
            return ClassificationLabel(label=data_object["label"], difficult=data_object["difficult"])
        if data_type == GroundTruthType.BOUNDING_BOX.value:
            top_left, bottom_right = data_object["points"]
            return BoundingBox(
                label=data_object["label"],
                top_left=top_left,
                bottom_right=bottom_right,
                difficult=data_object["difficult"],
            )
        if data_type == GroundTruthType.SEGMENTATION_MASK.value:
            return SegmentationMask(
                label=data_object["label"],
                points=data_object["points"],
                difficult=data_object["difficult"],
            )
        raise ValueError(f"invalid dictionary provided, unrecognized data type '{data_type}'")


# NOTE: intentionally leave undocumented; not a part of the public API
class ClassificationLabel(GroundTruth):
    label: str
    difficult: bool

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, label: str, difficult: bool = False):
        super().__init__()
        validate_label(label)
        self.label = label
        self.difficult = difficult

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=GroundTruthType.CLASSIFICATION_LABEL.value,
            data_object=dict(label=self.label, difficult=self.difficult),
        )


class BoundingBox(GroundTruth):
    """
    Ground truth data object representing a bounding box.

    Point coordinates should be in `(x, y)` format, as absolute pixel values.
    """

    label: str
    """Label to associate with this bounding box."""

    top_left: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the top left corner of the bounding box."""

    bottom_right: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the bottom right corner of the bounding box."""

    difficult: bool
    """
    A bounding box marked as `difficult` indicates that the object is considered difficult to recognize and should be
    ignored when evaluating metrics.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        label: str,
        top_left: Tuple[float, float],
        bottom_right: Tuple[float, float],
        difficult: bool = False,
    ):
        super().__init__()
        validate_label(label)
        self.label = label
        self.top_left = top_left
        self.bottom_right = bottom_right
        self.difficult = difficult

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=GroundTruthType.BOUNDING_BOX.value,
            data_object=dict(label=self.label, points=[self.top_left, self.bottom_right], difficult=self.difficult),
        )


class SegmentationMask(GroundTruth):
    """
    Ground truth data object representing a detection mask.

    Point coordinates should be in `(x, y)` format, as absolute pixel values.

    :raises ValueError: When fewer than three points are provided.
    """

    label: str
    """Label to associate with this segmentation mask."""

    points: List[Tuple[float, float]]
    """
    Polygon corresponding to the vertices of the segmentation mask. Must have at least three distinct elements, and
    may not cross itself or touch itself at any point.
    """

    difficult: bool
    """
    A segmentation mask marked as `difficult` indicates that the object is considered difficult to recognize and should
    be ignored when evaluating metrics.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, label: str, points: List[Tuple[float, float]], difficult: bool = False):
        super().__init__()
        validate_label(label)
        validate_polygon(points)
        self.label = label
        self.points = points
        self.difficult = difficult

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=GroundTruthType.SEGMENTATION_MASK.value,
            data_object=dict(label=self.label, points=self.points, difficult=self.difficult),
        )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from abc import abstractmethod
from collections import OrderedDict
from contextlib import contextmanager
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set

from deprecation import deprecated
from pydantic import validate_arguments

import kolena._api.v1.core as CoreAPI
from kolena._api.v1.detection import TestSuite as API
from kolena._api.v1.workflow import WorkflowType
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.consts import FieldName
from kolena._utils.endpoints import get_test_suite_url
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal import BaseTestCase
from kolena.errors import NotFoundError
from kolena.workflow._validators import assert_workflows_match


class BaseTestSuite(ABC, Frozen, WithTelemetry):
    """
    A test suite groups together one or more test cases.

    :param name: The name of the test suite to create or load.
    :param version: Optionally specify the version of the test suite to load. When absent, the latest version is loaded.
        Ignored when creating new test suites.
    :param description: Optionally specify a description for a newly created test suite. For existing test suites, this
        description can be edited via [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    :param test_cases: optionally specify a list of test cases to populate a new test suite. For existing test suites,
        test cases can be edited via [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
    """

    _id: int
    _workflow: WorkflowType

    @classmethod
    @abstractmethod
    def _test_case_from(cls, response: CoreAPI.TestCase.EntityData) -> BaseTestCase:
        pass

    def __init__(
        self,
        name: str,
        workflow: WorkflowType,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_cases: Optional[List[BaseTestCase]] = None,
        reset: bool = False,
    ):
        validate_name(name, FieldName.TEST_SUITE_NAME)
        self._validate_test_cases(test_cases)

        try:
            self._populate_from_other(self.load(name, version))
            if description is not None and self.description != description and not reset:
                log.warn("test suite already exists, not updating description when reset=False")
            if test_cases is not None:
                if self.version > 0 and not reset:
                    log.warn("test suite already exists, not updating test cases when reset=False")
                else:
                    self._hydrate(test_cases, description)
        except NotFoundError:
            if version is not None:
                log.warn(f"creating new test suite '{name}', ignoring provided version")
            self._populate_from_other(self._create(workflow, name, description, test_cases))
        self._freeze()

    @classmethod
    def _validate_test_cases(cls, test_cases: Optional[List[BaseTestCase]] = None) -> None:
        if test_cases:
            if any(cls._workflow != testcase._workflow for testcase in test_cases):
                raise TypeError("test case workflow does not match test suite's")

    def _populate_from_other(self, other: "BaseTestSuite") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self.test_cases = other.test_cases

    @classmethod
    def _create(
        cls,
        workflow: WorkflowType,
        name: str,
        description: Optional[str] = None,
        test_cases: Optional[List[BaseTestCase]] = None,
    ) -> "BaseTestSuite":
        """Create a new test suite with the provided name."""
        validate_name(name, FieldName.TEST_SUITE_NAME)
        request = CoreAPI.TestSuite.CreateRequest(name=name, description=description or "", workflow=workflow.value)
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.TestSuite.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        if test_cases is not None:
            obj._hydrate(test_cases)
        return obj

    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def _load_by_name(cls, name: str, version: Optional[int] = None) -> CoreAPI.TestSuite.EntityData:
        """Retrieve the existing test suite with the provided name."""
        request = CoreAPI.TestSuite.LoadByNameRequest(name=name, version=version)
        data = json.dumps(dataclasses.asdict(request))
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=data)
        krequests.raise_for_status(res)
        return from_dict(data_class=CoreAPI.TestSuite.EntityData, data=res.json())

    @validate_arguments(config=ValidatorConfig)
    def _hydrate(self, test_cases: List[BaseTestCase], description: Optional[str] = None) -> None:
        """Convenience method to hydrate an empty test suite with the provided test cases"""
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            if len(test_cases) > 0:
                for test_case in test_cases:
                    editor.add(test_case)

    @classmethod
    def _create_from_data(cls, data: CoreAPI.TestSuite.EntityData) -> "BaseTestSuite":
        assert_workflows_match(cls._workflow, data.workflow)
        obj = cls.__new__(cls)
        obj._id = data.id
        obj.name = data.name
        obj.version = data.version
        obj.description = data.description
        obj.test_cases = [cls._test_case_from(tc) for tc in data.test_cases]
        obj._freeze()
        return obj

    @classmethod
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        test_cases: Optional[List[BaseTestCase]] = None,
    ) -> "BaseTestSuite":
        """
        Create a new test suite with the provided name.

        :param name: The name of the new test suite to create.
        :param description: Optional free-form description of the test suite to create.
        :param test_cases: Optionally specify a set of test cases to populate the test suite.
        :return: The newly created test suite.
        """
        return cls._create(cls._workflow, name, description, test_cases)

    @classmethod
    def load(cls, name: str, version: Optional[int] = None) -> "BaseTestSuite":
        """
        Load an existing test suite with the provided name.

        :param name: The name of the test suite to load.
        :param version: Optionally specify a particular version of the test suite to load. Defaults to the latest
            version when unset.
        :return: The loaded test suite.
        """
        data = cls._load_by_name(name, version)
        obj = cls._create_from_data(data)
        log.info(f"loaded test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        return obj

    class Editor:
        """
        Interface to edit a test suite. Create with
        [`TestSuite.edit`][kolena.detection._internal.test_suite.BaseTestSuite.edit].
        """

        _test_cases: List[BaseTestCase]
        _reset: bool
        _description: str
        _initial_test_case_ids: Set[int]
        _initial_description: str

        _workflow: WorkflowType

        @validate_arguments(config=ValidatorConfig)
        def __init__(self, test_cases: List[BaseTestCase], description: str, reset: bool) -> None:
            self._test_cases: Dict[str, int] = OrderedDict()  # map from name -> id
            self._reset = reset
            self._description = description
            self._initial_test_case_ids = {tc._id for tc in test_cases}
            self._initial_description = description

        @validate_arguments(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """
            Update the description of the test suite.

            :param description: The new description of the test suite.
            """
            self._description = description

        @validate_arguments(config=ValidatorConfig)
        def add(self, test_case: BaseTestCase) -> None:
            """
            Add the provided [`TestCase`][kolena.detection.TestCase] to the test suite.
            If a different version of the test case already exists in this test suite, it is replaced.

            :param test_case: The test case to add to the test suite.
            """
            self._assert_workflows_match(test_case)
            self._test_cases[test_case.name] = test_case._id

        @validate_arguments(config=ValidatorConfig)
        def remove(self, test_case: BaseTestCase) -> None:
            """
            Remove the provided [`TestCase`][kolena.detection.TestCase] from the test suite. Any version of this test
            case in this test suite will be removed; the version does not need to match exactly.

            :param test_case: The test case to be removed from the test suite.
            """
            name = test_case.name
            if name not in self._test_cases.keys():
                raise KeyError(f"test case '{name}' not in test suite")
            self._test_cases.pop(name)

        @deprecated(details="use `TestSuite.Editor.add`", deprecated_in="0.56.0")
        @validate_arguments(config=ValidatorConfig)
        def merge(self, test_case: BaseTestCase) -> None:
            """
            !!! warning "Deprecated: since `0.56.0`"
                Replaced by idempotent behavior in
                [`TestSuite.Editor.add`][kolena.detection._internal.test_suite.BaseTestSuite.Editor.add].

            Add the provided [`TestCase`][kolena.detection.TestCase] to the test suite, replacing any previous version
            of the test case that may be present in the suite.

            :param test_case: The test case to be merged into the test suite.
            """
            self.add(test_case)

        def _assert_workflows_match(self, test_case: BaseTestCase) -> None:
            if test_case._workflow != self._workflow:
                raise ValueError(
                    f"test case workflow '{test_case._workflow}' mismatches test suite workflow '{self._workflow}'",
                )

        def _edited(self) -> bool:
            test_case_ids = set(self._test_cases.values())
            return self._description != self._initial_description or self._initial_test_case_ids != test_case_ids

    @contextmanager
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Edit this test suite in a context:

        ```python
        with test_suite.edit() as editor:
            # perform as many editing actions as desired
            editor.add(...)
            editor.remove(...)
        ```

        Changes are committed to Kolena when the context is exited.

        :param reset: Clear any and all test cases currently in the test suite.
        :return: Context-managed [`TestSuite.Editor`][kolena.detection._internal.test_suite.BaseTestSuite.Editor]
            instance exposing editing functionality for this test suite.
        """
        editor = BaseTestSuite.Editor(self.test_cases, self.description, reset)
        editor._workflow = self._workflow  # set outside of init such that parameter does not leak into documentation
        if not reset:
            editor.description(self.description)
            for test_case in self.test_cases:
                editor.add(test_case)

        yield editor

        if not editor._edited():
            log.info("no op: nothing edited")
            return

        log.info(f"editing test suite '{self.name}' (v{self.version})")
        request = CoreAPI.TestSuite.EditRequest(
            test_suite_id=self._id,
            current_version=self.version,
            name=self.name,
            description=editor._description,
            test_case_ids=list(editor._test_cases.values()),
        )
        data = json.dumps(dataclasses.asdict(request))
        res = krequests.post(endpoint_path=API.Path.EDIT.value, data=data)
        krequests.raise_for_status(res)
        test_suite_data = from_dict(data_class=CoreAPI.TestSuite.EntityData, data=res.json())
        log.success(f"edited test suite '{self.name}' (v{self.version}) ({get_test_suite_url(test_suite_data.id)})")
        with self._unfrozen():
            self.version = test_suite_data.version
            self.description = test_suite_data.description
            self.test_cases = [self._test_case_from(tc) for tc in test_suite_data.test_cases]
            self._id = test_suite_data.id



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from collections import defaultdict
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np

from kolena._utils import log
from kolena.classification.multiclass._utils import get_histogram_range
from kolena.classification.multiclass._utils import get_label_confidence
from kolena.classification.multiclass._utils import roc_curve
from kolena.classification.multiclass.workflow import AggregatedMetrics
from kolena.classification.multiclass.workflow import GroundTruth
from kolena.classification.multiclass.workflow import Inference
from kolena.classification.multiclass.workflow import InferenceLabel
from kolena.classification.multiclass.workflow import TestCase
from kolena.classification.multiclass.workflow import TestCaseMetrics
from kolena.classification.multiclass.workflow import TestSample
from kolena.classification.multiclass.workflow import TestSampleMetrics
from kolena.classification.multiclass.workflow import TestSuiteMetrics
from kolena.classification.multiclass.workflow import ThresholdConfiguration
from kolena.workflow import BarPlot
from kolena.workflow import ConfusionMatrix
from kolena.workflow import Curve
from kolena.workflow import CurvePlot
from kolena.workflow import EvaluationResults
from kolena.workflow import Histogram
from kolena.workflow import Plot
from kolena.workflow import TestCases
from kolena.workflow.annotation import ScoredClassificationLabel

Result = Tuple[TestSample, GroundTruth, Inference]


def _compute_test_sample_metric(
    threshold_configuration: ThresholdConfiguration,
    ground_truth: GroundTruth,
    inference: Inference,
) -> TestSampleMetrics:
    empty_metrics = TestSampleMetrics(
        classification=None,
        margin=None,
        is_correct=False,
    )
    if len(inference.inferences) == 0:
        return empty_metrics

    sorted_indices = np.argsort([label.score for label in inference.inferences])
    match = inference.inferences[sorted_indices[-1]]
    predicted_label, confidence_score = match.label, match.score
    margin: Optional[float] = None
    if len(sorted_indices) > 1:
        second_closest: Union[ScoredClassificationLabel, InferenceLabel] = inference.inferences[sorted_indices[-2]]
        margin = confidence_score - second_closest.score

    if threshold_configuration.threshold is not None and confidence_score < threshold_configuration.threshold:
        return empty_metrics

    is_correct = predicted_label == ground_truth.classification.label
    return TestSampleMetrics(
        classification=match,
        margin=margin,
        is_correct=is_correct,
    )


def _as_class_metric_plot(
    metric_name: str,
    metrics_by_label: Dict[str, AggregatedMetrics],
    labels: List[str],
) -> Optional[BarPlot]:
    if metric_name == "Recall":
        title = f"{metric_name} (TPR) vs. Class"
    else:
        title = f"{metric_name} vs. Class"

    values = [getattr(metrics_by_label[label], metric_name) for label in labels]
    valid_pairs = [(label, value) for label, value in zip(labels, values) if value != 0.0]
    if len(valid_pairs) > 0:
        return BarPlot(
            title=title,
            x_label="Class",
            y_label=metric_name,
            labels=[label for label, _ in valid_pairs],
            values=[value for _, value in valid_pairs],
        )
    return None


def _as_confidence_histogram(
    title: str,
    confidence_scores: List[float],
    confidence_range: Tuple[float, float, int] = (0, 1, 25),
) -> Histogram:
    min_range, max_range, bins = confidence_range
    hist, bins = np.histogram(
        confidence_scores,
        bins=bins,
        range=(min_range, max_range),
    )
    return Histogram(
        title=title,
        x_label="Confidence",
        y_label="Count",
        buckets=list(bins),
        frequency=list(hist),
    )


def _compute_confidence_histograms(
    test_case_name: str,
    metrics: List[TestSampleMetrics],
    confidence_range: Optional[Tuple[float, float, int]],
) -> List[Histogram]:
    if confidence_range is None:
        log.warn(
            f"skipping confidence histograms for {test_case_name}: unsupported confidence range",
        )
        return []

    confidence_all = [mts.classification.score for mts in metrics if mts.classification is not None]
    confidence_correct = [
        mts.classification.score for mts in metrics if mts.classification is not None and mts.is_correct
    ]
    confidence_incorrect = [
        mts.classification.score for mts in metrics if mts.classification is not None and not mts.is_correct
    ]

    plots = [
        _as_confidence_histogram("Score Distribution (All)", confidence_all, confidence_range),
        _as_confidence_histogram("Score Distribution (Correct)", confidence_correct, confidence_range),
        _as_confidence_histogram("Score Distribution (Incorrect)", confidence_incorrect, confidence_range),
    ]
    return plots


def _compute_test_case_plots(
    test_case_name: str,
    labels: List[str],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
    metrics: List[TestSampleMetrics],
    metrics_by_label: Dict[str, AggregatedMetrics],
    confidence_range: Optional[Tuple[float, float, int]],
) -> List[Plot]:
    gt_labels = {gt.classification.label for gt in ground_truths}
    plots: List[Plot] = [
        _as_class_metric_plot(field.name, metrics_by_label, labels)
        for field in dataclasses.fields(AggregatedMetrics)
        if len(gt_labels) > 2
        or field.name not in ["Precision", "Recall"]  # Omit single-class TC from precision and recall plots
    ]

    plots.extend(_compute_confidence_histograms(test_case_name, metrics, confidence_range))
    plots.append(_compute_test_case_ovr_roc_curve(test_case_name, gt_labels, ground_truths, inferences))
    plots.append(_compute_test_case_confusion_matrix(test_case_name, ground_truths, metrics))
    plots = list(filter(lambda plot: plot is not None, plots))

    return plots


def _compute_test_case_confusion_matrix(
    test_case_name: str,
    ground_truths: List[GroundTruth],
    metrics: List[TestSampleMetrics],
) -> Optional[Plot]:
    gt_labels: Set[str] = set()
    pred_labels: Set[str] = set()
    none_label = "None"
    # actual to predicted to count
    confusion_matrix: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
    for gt, metric in zip(ground_truths, metrics):
        actual_label = gt.classification.label
        predicted_label = metric.classification.label if metric.classification is not None else none_label
        gt_labels.add(actual_label)
        pred_labels.add(predicted_label)
        confusion_matrix[actual_label][predicted_label] += 1

    if len(gt_labels) < 2:
        log.warn(f"skipping confusion matrix for {test_case_name}: single label test case")
        return None

    labels: Set[str] = {*gt_labels, *pred_labels}
    contains_none = none_label in labels
    sortable_labels = [label for label in labels if label != none_label]
    ordered_labels = sorted(sortable_labels) if not contains_none else [*sorted(sortable_labels), none_label]
    matrix = []
    for actual_label in ordered_labels:
        matrix.append([confusion_matrix[actual_label][predicted_label] for predicted_label in ordered_labels])
    return ConfusionMatrix(title="Label Confusion Matrix", labels=ordered_labels, matrix=matrix)


def _compute_test_case_ovr_roc_curve(
    test_case_name: str,
    labels: List[str],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
) -> Optional[Plot]:
    if len(labels) > 10:
        log.warn(f"skipping one-vs-rest ROC curve for {test_case_name}: too many labels")
        return None

    curves = []
    for label in labels:
        y_true = [1 if gt.classification.label == label else 0 for gt in ground_truths]
        y_score = [get_label_confidence(label, inf.inferences) for inf in inferences]
        fpr_values, tpr_values = roc_curve(y_true, y_score)
        if len(fpr_values) > 0 and len(tpr_values) > 0:
            curves.append(Curve(x=fpr_values, y=tpr_values, label=label))

    if len(curves) > 0:
        return CurvePlot(
            title="Receiver Operating Characteristic (One-vs-Rest)",
            x_label="False Positive Rate (FPR)",
            y_label="True Positive Rate (TPR)",
            curves=curves,
        )
    return None


def _aggregate_label_metrics(
    labels: List[str],
    test_samples: List[TestSample],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
    metrics_test_samples: List[TestSampleMetrics],
) -> Dict[str, AggregatedMetrics]:
    aggregated_metrics = {}
    for base_label in labels:
        n_tp = 0
        n_fp = 0
        n_fn = 0
        n_tn = 0
        for ts, gt, inf, mts in zip(test_samples, ground_truths, inferences, metrics_test_samples):
            gt_label = gt.classification.label
            predicted_label = mts.classification.label if mts.classification is not None else None
            if gt_label == base_label and predicted_label == base_label:
                n_tp += 1
            if gt_label == base_label and predicted_label != base_label:
                n_fn += 1
            if gt_label != base_label and predicted_label == base_label:
                n_fp += 1
            if gt_label != base_label and predicted_label != base_label:
                n_tn += 1
        precision = n_tp / (n_tp + n_fp) if n_tp + n_fp > 0 else 0
        recall = n_tp / (n_tp + n_fn) if n_tp + n_fn > 0 else 0
        fpr = n_fp / (n_fp + n_tn) if n_fp + n_tn > 0 else 0
        f1_score = (2 * precision * recall) / (precision + recall) if precision + recall > 0 else 0
        aggregated_metrics[base_label] = AggregatedMetrics(
            F1=f1_score,
            Precision=precision,
            Recall=recall,
            FPR=fpr,
        )
    return aggregated_metrics


def _compute_test_case_metrics(
    test_samples: List[TestSample],
    ground_truths: List[GroundTruth],
    metrics_test_samples: List[TestSampleMetrics],
    metrics_by_label: Dict[str, AggregatedMetrics],
) -> TestCaseMetrics:
    n_correct = len([metric for metric in metrics_test_samples if metric.is_correct])
    n_images = len(test_samples)
    n_incorrect = n_images - n_correct
    accuracy = n_correct / n_images
    labels = {gt.classification.label for gt in ground_truths}

    macro_metrics_by_name: Dict[str, float] = {}
    for field in dataclasses.fields(AggregatedMetrics):
        metric_name = field.name
        metrics = [getattr(metrics_by_label[label], metric_name) for label in labels]
        macro_metrics_by_name[metric_name] = sum(metrics) / len(metrics)

    return TestCaseMetrics(
        n_labels=len(labels),
        n_correct=n_correct,
        n_incorrect=n_incorrect,
        accuracy=accuracy,
        macro_precision=macro_metrics_by_name["Precision"],
        macro_recall=macro_metrics_by_name["Recall"],
        macro_f1=macro_metrics_by_name["F1"],
        macro_tpr=macro_metrics_by_name["Recall"],
        macro_fpr=macro_metrics_by_name["FPR"],
    )


def _compute_test_suite_metrics(
    labels: List[str],
    test_samples: List[TestSample],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
    test_sample_metrics: List[TestSampleMetrics],
    test_case_metrics: List[TestCaseMetrics],
) -> TestSuiteMetrics:
    n_images = len(test_sample_metrics)
    n_correct = len([metric for metric in test_sample_metrics if metric.is_correct])
    mean_test_case_accuracy = sum([metric.accuracy for metric in test_case_metrics]) / len(test_case_metrics)

    values: Dict[str, Any] = dict(
        n_images=n_images,
        n_correct=n_correct,
        mean_test_case_accuracy=mean_test_case_accuracy,
    )
    fields: Dict[str, Type] = {}

    metrics_by_label = _aggregate_label_metrics(labels, test_samples, ground_truths, inferences, test_sample_metrics)
    for field in dataclasses.fields(AggregatedMetrics):
        attr = field.name
        label_values = [getattr(metric, attr) for metric in metrics_by_label.values()]
        mean_field_name = f"mean_{attr}"
        var_field_name = f"variance_{attr}"
        fields[mean_field_name] = float
        fields[var_field_name] = float
        values[mean_field_name] = np.mean(label_values)
        values[var_field_name] = np.var(label_values)

    dc = make_dataclass(
        "WorkflowTestSuiteMetrics",
        bases=(TestSuiteMetrics,),
        fields=list(fields.items()),
        frozen=True,
    )
    return dc(**values)


def MulticlassClassificationEvaluator(
    test_samples: List[TestSample],
    ground_truths: List[GroundTruth],
    inferences: List[Inference],
    test_cases: TestCases,
    configuration: ThresholdConfiguration = dataclasses.field(default_factory=ThresholdConfiguration),
) -> EvaluationResults:
    labels_set: Set[str] = set()
    for gt in ground_truths:
        labels_set.add(gt.classification.label)
    for inf in inferences:
        for label in inf.inferences:
            labels_set.add(label.label)
    labels = sorted(labels_set)

    metrics_test_sample: List[Tuple[TestSample, TestSampleMetrics]] = [
        (ts, _compute_test_sample_metric(configuration, gt, inf))
        for ts, gt, inf in zip(test_samples, ground_truths, inferences)
    ]
    test_sample_metrics: List[TestSampleMetrics] = [mts for _, mts in metrics_test_sample]
    confidence_scores = [mts.classification.score for mts in test_sample_metrics if mts.classification is not None]
    confidence_range = get_histogram_range(confidence_scores)

    metrics_test_case: List[Tuple[TestCase, TestCaseMetrics]] = []
    plots_test_case: List[Tuple[TestCase, List[Plot]]] = []
    for tc, tc_samples, tc_gts, tc_infs, tc_metrics in test_cases.iter(
        test_samples,
        ground_truths,
        inferences,
        test_sample_metrics,
    ):
        aggregated_label_metrics = _aggregate_label_metrics(labels, tc_samples, tc_gts, tc_infs, tc_metrics)
        test_case_metrics = _compute_test_case_metrics(tc_samples, tc_gts, tc_metrics, aggregated_label_metrics)
        metrics_test_case.append((tc, test_case_metrics))
        test_case_plots = _compute_test_case_plots(
            tc.name,
            labels,
            tc_gts,
            tc_infs,
            tc_metrics,
            aggregated_label_metrics,
            confidence_range,
        )
        plots_test_case.append((tc, test_case_plots))

    all_test_case_metrics = [metric for _, metric in metrics_test_case]
    metrics_test_suite = _compute_test_suite_metrics(
        labels,
        test_samples,
        ground_truths,
        inferences,
        test_sample_metrics,
        all_test_case_metrics,
    )

    return EvaluationResults(
        metrics_test_sample=metrics_test_sample,
        metrics_test_case=metrics_test_case,
        plots_test_case=plots_test_case,
        metrics_test_suite=metrics_test_suite,
    )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
import time
from abc import ABCMeta
from collections import defaultdict
from typing import Any
from typing import cast
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import pandas as pd
from pydantic import validate_arguments

from kolena._api.v1.generic import TestRun as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame_chunk
from kolena._utils.consts import BatchSize
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.endpoints import get_results_url
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import report_crash
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import ValidatorConfig
from kolena.errors import InputValidationError
from kolena.errors import WorkflowMismatchError
from kolena.workflow import Evaluator
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import GroundTruth
from kolena.workflow import Inference
from kolena.workflow import MetricsTestCase
from kolena.workflow import MetricsTestSample
from kolena.workflow import MetricsTestSuite
from kolena.workflow import Model
from kolena.workflow import Plot
from kolena.workflow import TestCase
from kolena.workflow import TestSample
from kolena.workflow import TestSuite
from kolena.workflow._datatypes import MetricsDataFrame
from kolena.workflow._datatypes import MetricsDataFrameSchema
from kolena.workflow._datatypes import TestSampleDataFrame
from kolena.workflow._datatypes import TestSampleDataFrameSchema
from kolena.workflow.evaluator import _configuration_description
from kolena.workflow.evaluator import _maybe_display_name
from kolena.workflow.evaluator import _maybe_evaluator_configuration_to_api
from kolena.workflow.evaluator_function import _is_configured
from kolena.workflow.evaluator_function import _TestCases
from kolena.workflow.evaluator_function import BasicEvaluatorFunction
from kolena.workflow.evaluator_function import EvaluationResults


class TestRun(Frozen, WithTelemetry, metaclass=ABCMeta):
    """
    A [`Model`][kolena.workflow.Model] tested on a [`TestSuite`][kolena.workflow.TestSuite] using a specific
    [`Evaluator`][kolena.workflow.Evaluator] implementation.

    :param model: The model being tested.
    :param test_suite: The test suite on which to test the model.
    :param evaluator: An optional evaluator implementation.
        Requires a previously configured server-side evaluator to default to if omitted.
        (Please see [`BasicEvaluatorFunction`][kolena.workflow.evaluator_function.BasicEvaluatorFunction] for type
        definition.)
    :param configurations: a list of configurations to use when running the evaluator.
    :param reset: overwrites existing inferences if set.
    """

    _id: int

    model: Model
    test_suite: TestSuite
    evaluator: Union[Evaluator, BasicEvaluatorFunction, None]
    configurations: Optional[List[EvaluatorConfiguration]]

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        model: Model,
        test_suite: TestSuite,
        evaluator: Union[Evaluator, BasicEvaluatorFunction, None] = None,
        configurations: Optional[List[EvaluatorConfiguration]] = None,
        reset: bool = False,
        dry_run: bool = False,
    ):
        if configurations is None:
            configurations = []

        if model.workflow != test_suite.workflow:
            raise WorkflowMismatchError(
                f"model workflow ({model.workflow}) does not match test suite workflow ({test_suite.workflow})",
            )

        if reset:
            log.warn("overwriting existing inferences from this model (reset=True)")
        else:
            log.info("not overwriting any existing inferences from this model (reset=False)")

        self.model = model
        self.test_suite = test_suite
        self.evaluator = evaluator
        self.configurations = self.evaluator.configurations if isinstance(evaluator, Evaluator) else configurations
        self.reset = reset
        self.dry_run = dry_run

        evaluator_display_name = (
            None
            if evaluator is None
            else evaluator.display_name()
            if isinstance(evaluator, Evaluator)
            else evaluator.__name__
        )
        api_configurations = (
            [_maybe_evaluator_configuration_to_api(config) for config in self.configurations]
            if self.configurations is not None
            else None
        )

        request = API.CreateOrRetrieveRequest(
            model_id=model._id,
            test_suite_id=test_suite._id,
            evaluator=evaluator_display_name,
            configurations=api_configurations,
        )
        res = krequests.put(
            endpoint_path=API.Path.CREATE_OR_RETRIEVE.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)
        response = from_dict(data_class=API.CreateOrRetrieveResponse, data=res.json())
        self._id = response.test_run_id
        self._freeze()

    def run(self) -> Dict[str, Any]:
        """
        Run the testing process, first extracting inferences for all test samples in the test suite then performing
        evaluation.
        """
        try:
            inferences = []
            for ts in log.progress_bar(self.iter_test_samples(), desc="performing inference"):
                if self.model.infer is None:  # only fail when `infer` is necessary
                    raise ValueError("model must implement `infer`")
                inferences.append((ts, self.model.infer(ts)))

            if len(inferences) > 0:
                log.success(f"performed inference on {len(inferences)} test samples")
                if not self.dry_run:
                    log.info("uploading inferences")
                    self.upload_inferences(inferences)

            metrics = self.evaluate()
            return dict(inferences=inferences, metrics=metrics)
        except Exception as e:
            report_crash(self._id, API.Path.MARK_CRASHED.value)
            raise e

    def load_test_samples(self) -> List[TestSample]:
        """
        Load the test samples in the test suite that do not yet have inferences uploaded.

        :return: a list of all test samples in the test suite still requiring inferences.
        """
        return list(self.iter_test_samples())

    def iter_test_samples(self) -> Iterator[TestSample]:
        """
        Iterate through the test samples in the test suite that do not yet have inferences uploaded.

        :return: an iterator over each test sample still requiring inferences.
        """
        test_sample_type = self.model.workflow.test_sample_type
        for df_batch in self._iter_test_samples_batch():
            for record in df_batch.itertuples():
                yield test_sample_type._from_dict(record.test_sample)

    def _iter_all_inferences(self) -> Iterator[Tuple[TestSample, GroundTruth, Inference]]:
        """
        Iterate over all inferences stored for this model on the provided test case.

        :return: an iterator exposing the ground truths and inferences for all test samples in the test run.
        """
        log.info(f"loading inferences from model '{self.model.name}' on test suite '{self.test_suite.name}'")
        for df_batch in _BatchedLoader.iter_data(
            init_request=API.LoadTestSampleInferencesRequest(
                test_run_id=self._id,
                batch_size=BatchSize.LOAD_SAMPLES.value,
            ),
            endpoint_path=API.Path.LOAD_INFERENCES.value,
            df_class=TestSampleDataFrame,
        ):
            for record in df_batch.itertuples():
                test_sample = self.test_suite.workflow.test_sample_type._from_dict(record.test_sample)
                ground_truth = self.test_suite.workflow.ground_truth_type._from_dict(record.ground_truth)
                inference = self.test_suite.workflow.inference_type._from_dict(record.inference)
                yield test_sample, ground_truth, inference
        log.info(f"loaded inferences from model '{self.model.name}' on test suite '{self.test_suite.name}'")

    @validate_arguments(config=ValidatorConfig)
    def upload_inferences(self, inferences: List[Tuple[TestSample, Inference]]) -> None:
        """
        Upload inferences from a model.

        :param inferences: the inferences, paired with their corresponding test samples, to upload.
        """
        if len(inferences) == 0:
            return

        inference_dicts = [(ts._to_dict(), inf._to_dict()) for ts, inf in inferences]
        df = pd.DataFrame(inference_dicts, columns=["test_sample", "inference"])
        df_validated = TestSampleDataFrame(validate_df_schema(df, TestSampleDataFrameSchema, trusted=True))
        df_serializable = df_validated.as_serializable()

        init_response = init_upload()
        upload_data_frame_chunk(df_serializable, init_response.uuid)

        request = API.UploadInferencesRequest(uuid=init_response.uuid, test_run_id=self._id, reset=self.reset)
        res = krequests.put(
            endpoint_path=API.Path.UPLOAD_INFERENCES.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)

    def evaluate(self) -> Dict[str, Any]:
        """
        Perform evaluation by computing metrics for individual test samples, in aggregate across test cases, and across
        the complete test suite at each [`EvaluatorConfiguration`][kolena.workflow.EvaluatorConfiguration].
        """
        if self.evaluator is None and not self.dry_run:
            log.info("commencing server side metrics evaluation")
            self._start_server_side_evaluation()
            return {}

        # TODO: assert that testing is complete?
        t0 = time.time()
        log.info("commencing evaluation")
        if isinstance(self.evaluator, Evaluator):
            metrics = self._perform_evaluation(self.evaluator)
        else:
            metrics = self._perform_streamlined_evaluation(self.evaluator)

        log.success(f"completed evaluation in {time.time() - t0:0.1f} seconds")
        log.success(f"results: {get_results_url(self.model.workflow.name, self.model._id, self.test_suite._id)}")

        return metrics

    def _perform_evaluation(self, evaluator: Evaluator) -> Dict[str, Any]:
        configurations: Sequence[Optional[EvaluatorConfiguration]] = (
            cast(Sequence[Optional[EvaluatorConfiguration]], evaluator.configurations)
            if len(evaluator.configurations) > 0
            else [None]
        )
        test_case_metrics: Dict[int, Dict[Optional[EvaluatorConfiguration], MetricsTestCase]] = {}
        test_case_plots: Dict[int, Dict[Optional[EvaluatorConfiguration], Optional[List[Plot]]]] = {}

        for test_case in self.test_suite.test_cases:
            log.info(f"evaluating test case '{test_case.name}'")
            test_case_metrics_by_config = {}
            test_case_plots_by_config = {}
            inferences = self.model.load_inferences(test_case)

            for configuration in configurations:
                configuration_description = _configuration_description(configuration)
                log.info(f"computing test sample metrics {configuration_description}")
                metrics_test_sample = evaluator.compute_test_sample_metrics(test_case, inferences, configuration)

                if not self.dry_run:
                    log.info(f"uploading test sample metrics {configuration_description}")
                    self._upload_test_sample_metrics(test_case, metrics_test_sample, configuration)

                log.info(f"computing test case metrics {configuration_description}")
                # TODO: sort? order returned from evaluator may not match inferences order
                mts = [metrics for _, metrics in metrics_test_sample]
                metrics_test_case = evaluator.compute_test_case_metrics(test_case, inferences, mts, configuration)
                test_case_metrics_by_config[configuration] = metrics_test_case

                log.info(f"computing test case plots {configuration_description}")
                plots_test_case = evaluator.compute_test_case_plots(test_case, inferences, mts, configuration)
                test_case_plots_by_config[configuration] = plots_test_case

            test_case_metrics[test_case._id] = test_case_metrics_by_config
            test_case_plots[test_case._id] = test_case_plots_by_config

        if not self.dry_run:
            log.info("uploading test case metrics")
            self._upload_test_case_metrics(test_case_metrics)
            log.info("uploading test case plots")
            self._upload_test_case_plots(test_case_plots)

        log.info("computing test suite metrics")
        test_suite_metrics: Dict[Optional[EvaluatorConfiguration], Optional[MetricsTestSuite]] = {}
        for configuration in configurations:
            test_case_with_metrics = [
                (tc, test_case_metrics[tc._id][configuration]) for tc in self.test_suite.test_cases
            ]
            log.info(f"computing test suite metrics {_configuration_description(configuration)}")
            metrics_test_suite = evaluator.compute_test_suite_metrics(
                self.test_suite,
                test_case_with_metrics,
                configuration,
            )
            test_suite_metrics[configuration] = metrics_test_suite

        if not self.dry_run:
            log.info("uploading test suite metrics")
            self._upload_test_suite_metrics(test_suite_metrics)

        return dict(
            test_case_metrics=test_case_metrics,
            test_case_plots=test_case_plots,
            test_suite_metrics=test_suite_metrics,
        )

    def _perform_streamlined_evaluation(self, evaluator: BasicEvaluatorFunction) -> Dict[str, Any]:
        test_samples, ground_truths, inferences = [], [], []
        for sample, gt, inf in self._iter_all_inferences():
            test_samples.append(sample)
            ground_truths.append(gt)
            inferences.append(inf)
        test_case_membership: List[Tuple[TestCase, List[TestSample]]] = self.test_suite.load_test_samples()
        test_case_test_samples = _TestCases(
            test_case_membership,
            self._id,
            len(self.configurations),
        )

        test_case_metrics: Dict[int, Dict[Optional[EvaluatorConfiguration], MetricsTestCase]] = defaultdict(dict)
        test_case_plots: Dict[int, Dict[Optional[EvaluatorConfiguration], List[Plot]]] = defaultdict(dict)
        test_suite_metrics: Dict[Optional[EvaluatorConfiguration], MetricsTestSuite] = dict()

        def process_results(results: Optional[EvaluationResults], config: Optional[EvaluatorConfiguration]) -> None:
            if results is None:
                log.info(f"no results {_configuration_description(config)}")
                return

            if not self.dry_run:
                log.info(f"uploading test sample metrics {_configuration_description(config)}")
                self._upload_test_sample_metrics(
                    test_case=None,
                    metrics=results.metrics_test_sample,
                    configuration=config,
                )
            for test_case, metrics in results.metrics_test_case:
                test_case_metrics[test_case._id][config] = metrics
            for test_case, plots in results.plots_test_case:
                test_case_plots[test_case._id][config] = plots
            if results.metrics_test_suite is not None:
                test_suite_metrics[config] = results.metrics_test_suite

        if _is_configured(evaluator):
            for configuration in self.configurations:
                test_case_test_samples._set_configuration(configuration)
                evaluation_results = evaluator(
                    test_samples,
                    ground_truths,
                    inferences,
                    test_case_test_samples,
                    configuration,
                )
                process_results(evaluation_results, configuration)
        else:
            test_case_test_samples._set_configuration(None)
            evaluation_results = evaluator(test_samples, ground_truths, inferences, test_case_test_samples)
            process_results(evaluation_results, None)

        if not self.dry_run:
            log.info("uploading test case metrics")
            self._upload_test_case_metrics(test_case_metrics)
            log.info("uploading test case plots")
            self._upload_test_case_plots(test_case_plots)
            log.info("uploading test suite metrics")
            self._upload_test_suite_metrics(test_suite_metrics)

        return dict(
            test_case_metrics=test_case_metrics,
            test_case_plots=test_case_plots,
            test_suite_metrics=test_suite_metrics,
        )

    def _iter_test_samples_batch(
        self,
        batch_size: int = BatchSize.LOAD_SAMPLES.value,
    ) -> Iterator[TestSampleDataFrame]:
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")
        init_request = API.LoadRemainingTestSamplesRequest(
            test_run_id=self._id,
            batch_size=batch_size,
            load_all=self.reset,
        )
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.LOAD_TEST_SAMPLES.value,
            df_class=TestSampleDataFrame,
        )

    @validate_arguments(config=ValidatorConfig)
    def _upload_test_sample_metrics(
        self,
        test_case: Optional[TestCase],
        metrics: List[Tuple[TestSample, MetricsTestSample]],
        configuration: Optional[EvaluatorConfiguration],
    ) -> None:
        metrics_records = [(ts._to_dict(), ts_metrics._to_dict()) for ts, ts_metrics in metrics]
        df = pd.DataFrame(metrics_records, columns=["test_sample", "metrics"])
        df_validated = MetricsDataFrame(validate_df_schema(df, MetricsDataFrameSchema, trusted=True))
        df_serializable = df_validated.as_serializable()

        init_response = init_upload()
        upload_data_frame_chunk(df_serializable, init_response.uuid)

        request = API.UploadTestSampleMetricsRequest(
            uuid=init_response.uuid,
            test_run_id=self._id,
            test_case_id=test_case._id if test_case is not None else None,
            configuration=_maybe_evaluator_configuration_to_api(configuration),
        )
        res = krequests.put(
            endpoint_path=API.Path.UPLOAD_TEST_SAMPLE_METRICS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)

    def _upload_test_case_metrics(
        self,
        metrics: Dict[int, Dict[Optional[EvaluatorConfiguration], MetricsTestCase]],
    ) -> None:
        records = [
            (test_case_id, _maybe_display_name(config), tc_metrics._to_dict())
            for test_case_id, tc_metrics_by_config in metrics.items()
            for config, tc_metrics in tc_metrics_by_config.items()
        ]
        df = pd.DataFrame(records, columns=["test_case_id", "configuration_display_name", "metrics"])
        return self._upload_aggregate_metrics(API.Path.UPLOAD_TEST_CASE_METRICS.value, df)

    def _upload_test_case_plots(
        self,
        plots: Dict[int, Dict[Optional[EvaluatorConfiguration], Optional[List[Plot]]]],
    ) -> None:
        records = [
            (test_case_id, _maybe_display_name(config), tc_plot._to_dict())
            for test_case_id, tc_plots_by_config in plots.items()
            for config, tc_plots in tc_plots_by_config.items()
            for tc_plot in tc_plots or []
        ]
        df = pd.DataFrame(records, columns=["test_case_id", "configuration_display_name", "metrics"])
        return self._upload_aggregate_metrics(API.Path.UPLOAD_TEST_CASE_PLOTS.value, df)

    def _upload_test_suite_metrics(
        self,
        metrics: Dict[Optional[EvaluatorConfiguration], Optional[MetricsTestSuite]],
    ) -> None:
        records: List[Tuple[Optional[str], Dict[str, Any]]] = [
            (_maybe_display_name(config), ts_metrics._to_dict())
            for config, ts_metrics in metrics.items()
            if ts_metrics is not None
        ]
        df = pd.DataFrame(records, columns=["configuration_display_name", "metrics"])
        return self._upload_aggregate_metrics(API.Path.UPLOAD_TEST_SUITE_METRICS.value, df)

    def _upload_aggregate_metrics(self, endpoint_path: str, df: pd.DataFrame) -> None:
        df_validated = MetricsDataFrame(validate_df_schema(df, MetricsDataFrameSchema, trusted=True))
        df_serializable = df_validated.as_serializable()

        init_response = init_upload()
        upload_data_frame_chunk(df_serializable, init_response.uuid)

        request = API.UploadAggregateMetricsRequest(
            uuid=init_response.uuid,
            test_run_id=self._id,
            test_suite_id=self.test_suite._id,
        )
        res = krequests.put(endpoint_path=endpoint_path, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)

    def _start_server_side_evaluation(self) -> None:
        request = API.EvaluateRequest(test_run_id=self._id)
        res = krequests.put(endpoint_path=API.Path.EVALUATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)


@validate_arguments(config=ValidatorConfig)
def test(
    model: Model,
    test_suite: TestSuite,
    evaluator: Union[Evaluator, BasicEvaluatorFunction, None] = None,
    configurations: Optional[List[EvaluatorConfiguration]] = None,
    reset: bool = False,
    dry_run: bool = False,
) -> Dict[str, Any]:
    """
    Test a [`Model`][kolena.workflow.Model] on a [`TestSuite`][kolena.workflow.TestSuite] using a specific
    [`Evaluator`][kolena.workflow.Evaluator] implementation.

    ```python
    from kolena.workflow import test

    test(model, test_suite, evaluator, reset=True)
    ```

    :param model: The model being tested, implementing the `infer` method.
    :param test_suite: The test suite on which to test the model.
    :param evaluator: An optional evaluator implementation.
        Requires a previously configured server-side evaluator to default to if omitted.
        (Please see [`BasicEvaluatorFunction`][kolena.workflow.evaluator_function.BasicEvaluatorFunction] for type
        definition.)
    :param configurations: A list of configurations to use when running the evaluator.
    :param reset: Overwrites existing inferences if set.
    """
    return TestRun(model, test_suite, evaluator, configurations, reset, dry_run).run()



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test samples are the inputs to your models when testing.

For example, for a model that processes specific regions within a larger image, its test sample may be defined:

```python
from dataclasses import dataclass

from kolena.workflow import Image
from kolena.workflow.annotation import BoundingBox

@dataclass(frozen=True)
class ImageWithRegion(Image):
    region: BoundingBox

example = ImageWithRegion(
    locator="s3://my-bucket/example-image.png",  # field from Image base class
    region=BoundingBox(top_left=(0, 0), bottom_right=(100, 100)),
)
```
"""
import copy
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from pydantic import StrictBool
from pydantic import StrictFloat
from pydantic import StrictInt
from pydantic import StrictStr
from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow._datatypes import DataType
from kolena.workflow._datatypes import TypedDataObject
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import safe_issubclass
from kolena.workflow._validators import validate_field
from kolena.workflow._validators import validate_metadata_dict
from kolena.workflow.asset import ImageAsset

Metadata = Dict[
    str,
    Union[
        None,
        # prevent coercion of values in metadata -- see: https://pydantic-docs.helpmanual.io/usage/types/#strict-types
        StrictStr,
        StrictFloat,
        StrictInt,
        StrictBool,
        # Pydantic's StrictX doesn't play nicely with deserialization (e.g. isinstance("a string", StrictStr) => False)
        #  -- include base scalar types as fallbacks for this purpose
        str,
        float,
        int,
        bool,
        List[
            Union[
                None,
                StrictStr,
                StrictFloat,
                StrictInt,
                StrictBool,
                str,
                float,
                int,
                bool,
            ]
        ],
    ],
]
"""
Type of the `metadata` field that can be included on [`TestSample`][kolena.workflow.TestSample] definitions. String
(`str`) keys and scalar values (`int`, `float`, `str`, `bool`, `None`) as well as scalar list values are permitted.

```python
from dataclasses import dataclass, field
from kolena.workflow import Image, Metadata

@dataclass(frozen=True)
class ImageWithMetadata(Image):
    metadata: Metadata = field(default_factory=dict)
```
"""

_METADATA_KEY = "metadata"


class _TestSampleType(DataType):
    IMAGE = "IMAGE"
    TEXT = "TEXT"
    VIDEO = "VIDEO"
    DOCUMENT = "DOCUMENT"
    COMPOSITE = "COMPOSITE"
    POINT_CLOUD = "POINT_CLOUD"
    CUSTOM = "CUSTOM"

    @staticmethod
    def _data_category() -> str:
        return "TEST_SAMPLE"


@dataclass(frozen=True, config=ValidatorConfig)
class TestSample(TypedDataObject[_TestSampleType], metaclass=ABCMeta):
    """
    The inputs to a model.

    Test samples can be customized as necessary for a workflow by extending this class or one of the built-in test
    sample types.

    Extensions to the `TestSample` class may define a `metadata` field of type
    [`Metadata`][kolena.workflow.test_sample.Metadata] containing a dictionary of scalar properties associated with the
    test sample, intended for use when sorting or filtering test samples.

    Kolena handles the `metadata` field differently from other test sample fields. Updates to the `metadata` object
    for a given test sample are merged with previously uploaded metadata. As such, `metadata` for a given test sample
    within a test case is **not** immutable, and should **not** be relied on when an implementation of
    [`Model`][kolena.workflow.Model] computes inferences, or when an implementation of
    [`Evaluator`][kolena.workflow.Evaluator] evaluates metrics.
    """

    @staticmethod
    def _data_type() -> _TestSampleType:
        return _TestSampleType.CUSTOM

    def _to_dict(self) -> Dict[str, Any]:
        base_dict = super()._to_dict()
        base_dict.pop(_METADATA_KEY, None)
        return base_dict

    def _to_metadata_dict(self) -> Metadata:
        base_dict = super()._to_dict()
        return base_dict.pop(_METADATA_KEY, {})


@dataclass(frozen=True, config=ValidatorConfig)
class Composite(TestSample):
    """
    A test sample composed of multiple basic [`TestSample`][kolena.workflow.TestSample] elements.

    An example application would be each test sample is a pair of face images, and the goal is to predict whether the
    two images are of the same person. For this use-case the test sample can be defined as:

    ```python
    class FacePairSample(Composite):
        source: Image
        target: Image
    ```

    To facilitate visualization for this kind of use cases, see usage of [`GroundTruth`][kolena.workflow.GroundTruth]
    and [`Inference`][kolena.workflow.Inference].
    """

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.COMPOSITE


@dataclass(frozen=True, config=ValidatorConfig)
class Image(TestSample):
    """An image located in a cloud bucket or served at a URL."""

    locator: str
    """The URL of this image, using e.g. `s3`, `gs`, or `https` scheme (`s3://my-bucket/path/to/image.png`)."""

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.IMAGE


@dataclass(frozen=True, config=ValidatorConfig)
class ImagePair(Composite):
    """Two [`Image`s][kolena.workflow.Image] paired together."""

    a: Image
    """The left [`Image`][kolena.workflow.Image] in the image pair."""

    b: Image
    """The right [`Image`][kolena.workflow.Image] in the image pair."""


@dataclass(frozen=True, config=ValidatorConfig)
class Text(TestSample):
    """An inline text snippet."""

    text: str
    """The text snippet."""

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.TEXT


@dataclass(frozen=True, config=ValidatorConfig)
class ImageText(Composite):
    """An image paired with a text snippet."""

    image: Image
    """The [`Image`][kolena.workflow.Image] in this image-text pair."""

    text: Text
    """The text snippet in this image-text pair."""


# NOTE: declare BaseVideo as separate class for extension -- default fields in main Video class prevent extension with
#  non-default fields
@dataclass(frozen=True, config=ValidatorConfig)
class BaseVideo(TestSample):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.VIDEO


@dataclass(frozen=True, config=ValidatorConfig)
class Video(BaseVideo):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    thumbnail: Optional[ImageAsset] = None
    """Optionally provide asset locator for custom video thumbnail."""

    start: Optional[float] = None
    """Optionally specify start time of video snippet, in seconds."""

    end: Optional[float] = None
    """Optionally specify end time of video snippet, in seconds."""

    def __post_init__(self) -> None:
        if self.start is not None and self.end is not None and self.start > self.end:
            raise ValueError(f"Specified start time '{self.start}' is after specified end time '{self.end}'")
        if self.start is not None and self.end is not None and (self.start < 0 or self.end < 0):
            raise ValueError(f"Specified start time '{self.start}' and end time '{self.end}' must be non-negative")


@dataclass(frozen=True, config=ValidatorConfig)
class Document(TestSample):
    """A remotely linked document, e.g. PDF or TXT file."""

    locator: str
    """URL (e.g. S3, HTTPS) of the document."""

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.DOCUMENT


@dataclass(frozen=True, config=ValidatorConfig)
class PointCloud(TestSample):
    """A pointcloud file located in a cloud bucket or served at a URL."""

    locator: str
    """The URL of the pointcloud file, using e.g. `s3`, `gs`, or `https` scheme (`s3://my-bucket/path/to/image.pcd`)."""

    @classmethod
    def _data_type(cls) -> _TestSampleType:
        return _TestSampleType.POINT_CLOUD


_TEST_SAMPLE_BASE_TYPES = [Composite, Image, Text, BaseVideo, Document, PointCloud]


def _validate_test_sample_type(test_sample_type: Type[TestSample], recurse: bool = True) -> None:
    if not issubclass(test_sample_type, TestSample):
        raise ValueError(f"Test sample must subclass {TestSample.__name__}")

    if (
        not issubclass(test_sample_type, tuple(_TEST_SAMPLE_BASE_TYPES))
        and test_sample_type._data_type() is not _TestSampleType.CUSTOM
    ):
        supported_bases = ", ".join(t.__name__ for t in _TEST_SAMPLE_BASE_TYPES)
        raise ValueError(
            f"Test sample not extending a supported base must declare data_type {_TestSampleType.CUSTOM.value}.\n"
            f"Supported bases: {supported_bases}",
        )

    # TODO: this is structurally identical to implementation in validate_ground_truth_type and
    #  validate_metrics_test_sample_type -- share?
    fields_by_name = copy.copy(get_data_object_field_types(test_sample_type))
    is_composite = issubclass(test_sample_type, Composite)
    for field_name, field_value in fields_by_name.items():
        if field_name == _METADATA_KEY:
            validate_metadata_dict(field_value)
        # check composite test sample, keeping recurse to restrict 1-level composition for now
        elif is_composite and safe_issubclass(field_value, TestSample):
            if not recurse:
                raise ValueError(f"Nested composite test sample is not supported: {field_name}.")
            _validate_test_sample_type(field_value, recurse=False)
        else:
            validate_field(field_name, field_value)


def _get_composite_fields(test_sample_type: Type[TestSample]) -> List[str]:
    composite_fields = []
    for k, v in get_data_object_field_types(test_sample_type).items():
        if k != _METADATA_KEY:
            if safe_issubclass(v, TestSample):
                composite_fields.append(k)

    return composite_fields



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from typing import List
from typing import Tuple

from deprecation import deprecated
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass
from tqdm import tqdm

from kolena._api.v1.batched_load import BatchedLoad as LoadAPI
from kolena._api.v1.fr import Asset as AssetAPI
from kolena._api.v1.fr import TestRun as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.asset_path_mapper import AssetPathMapper
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.dataframes.validators import validate_df_record_count
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import report_crash
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import ValidatorConfig
from kolena.errors import InputValidationError
from kolena.fr import InferenceModel
from kolena.fr import Model
from kolena.fr import TestSuite
from kolena.fr._utils import upload_image_chips
from kolena.fr.datatypes import _ImageChipsDataFrame
from kolena.fr.datatypes import _ResultStageFrame
from kolena.fr.datatypes import EmbeddingDataFrame
from kolena.fr.datatypes import ImageDataFrame
from kolena.fr.datatypes import ImageResultDataFrame
from kolena.fr.datatypes import ImageResultDataFrameSchema
from kolena.fr.datatypes import PairDataFrame
from kolena.fr.datatypes import PairResultDataFrame
from kolena.fr.datatypes import PairResultDataFrameSchema


class TestRun(ABC, Frozen, WithTelemetry):
    """
    Interface to test a [`Model`][kolena.fr.Model] on a [`TestSuite`][kolena.fr.TestSuite]. Any in-progress tests for
    this model on this test suite is resumed.

    For a streamlined interface, see [`test`][kolena.fr.test].

    :param model: The model being tested.
    :param test_suite: The test suite on which to test the model.
    :param reset: Overwrites existing inferences if set.
    """

    _id: int

    @dataclass(frozen=True, config=ValidatorConfig)
    class Data:
        id: int
        model: Model.Data
        test_suites: List[TestSuite.Data]

    def __init__(self, model: Model, test_suite: TestSuite, reset: bool = False):
        if reset:
            log.warn("overwriting existing inferences from this model (reset=True)")
        else:
            log.info("not overwriting any existing inferences from this model (reset=False)")

        request = API.CreateOrRetrieveRequest(model_id=model.data.id, test_suite_ids=[test_suite._id], reset=reset)
        res = krequests.post(
            endpoint_path=API.Path.CREATE_OR_RETRIEVE.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(res)
        response = from_dict(data_class=TestRun.Data, data=res.json())

        self.data = response
        self._id = response.id
        self._model = model
        self._test_suite = test_suite
        self._reset = reset
        self._freeze()

    @classmethod
    @deprecated(details="use `TestRun.__init__` directly", deprecated_in="0.58.0")
    def create_or_retrieve(cls, model: Model, test_suite: TestSuite, reset: bool = False) -> "TestRun":
        """
        !!! warning "Deprecated: since `0.57.0`"
            Use the [`TestRun`][kolena.fr.TestRun] constructor instead.

        Create a new test run for the provided [`Model`][kolena.fr.Model] on the provided
        [`TestSuite`][kolena.fr.TestSuite]. If a test run for this model on this suite already exists, it is returned.

        :param model: The model being tested.
        :param test_suite: The test suite on which to test the model.
        :param reset: Overwrites existing inferences if set.
        :return: The created or retrieved test run.
        """
        return TestRun(model, test_suite, reset=reset)

    @validate_arguments
    def load_remaining_images(self, batch_size: int = 10_000_000) -> ImageDataFrame:
        """
        Load a DataFrame containing records for each of the images in the configured test suite that does not yet have
        results from the configured model.

        :param batch_size: Optionally specify the maximum number of image records to return.
        :return: DataFrame containing records for each of the images that must be processed.
        :raises InputValidationError: The requested `batch_size` failed validation.
        :raises RemoteError: Images could not be loaded for any reason.
        """
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")
        log.info("loading remaining images for test run")
        init_request = API.InitLoadRemainingImagesRequest(
            test_run_id=self.data.id,
            batch_size=batch_size,
            load_all=self._reset,
        )
        with krequests.put(
            endpoint_path=API.Path.INIT_LOAD_REMAINING_IMAGES.value,
            data=json.dumps(dataclasses.asdict(init_request)),
            stream=True,
        ) as init_res:
            krequests.raise_for_status(init_res)

            load_uuid = None
            try:
                dfs: List[ImageDataFrame] = []
                for line in init_res.iter_lines():
                    partial_response = from_dict(data_class=LoadAPI.InitDownloadPartialResponse, data=json.loads(line))
                    load_uuid = partial_response.uuid
                    dfs.append(_BatchedLoader.load_path(partial_response.path, ImageDataFrame))
                log.info("loaded remaining images for test run")
                return _BatchedLoader.concat(dfs, ImageDataFrame)
            finally:
                _BatchedLoader.complete_load(load_uuid)

    def upload_image_results(self, df_image_result: ImageResultDataFrame) -> int:
        """
        Upload inference results for a batch of images.

        All columns except for `image_id` and `embedding` are optional. An empty `embedding` cell in a record
        indicates a failure to enroll. The `failure_reason` column can optionally be specified for failures to enroll.

        To provide more than one embedding extracted from a given image, include multiple records with the same
        `image_id` in `df_image_result` (one for each embedding extracted). Records for a given `image_id` must
        be submitted in the same `df_image_result` DataFrame, and **not** across multiple calls to
        `upload_image_results`.

        :param df_image_result: DataFrame of any size containing records describing inference results for an image.
        :return: Number of records successfully uploaded.
        :raises TypeValidationError: The DataFrame failed type validation.
        :raises RemoteError: The DataFrame was unable to be successfully ingested for any reason.
        """
        log.info("uploading inference results for test run")
        init_response = init_upload()

        asset_config_res = krequests.get(endpoint_path=AssetAPI.Path.CONFIG.value)
        krequests.raise_for_status(asset_config_res)
        asset_config = from_dict(data_class=AssetAPI.Config, data=asset_config_res.json())
        asset_path_mapper = AssetPathMapper(asset_config)

        df_validated = ImageResultDataFrame(validate_df_schema(df_image_result, ImageResultDataFrameSchema))
        validate_df_record_count(df_validated)
        df_image_chips = _ImageChipsDataFrame.from_image_result_data_frame(
            test_run_id=self.data.id,
            load_uuid=init_response.uuid,
            df=df_validated,
        )
        upload_image_chips(df_image_chips)
        df_result_stage = _ResultStageFrame.from_image_result_data_frame(
            test_run_id=self.data.id,
            load_uuid=init_response.uuid,
            df=df_validated,
            path_mapper=asset_path_mapper,
        )
        upload_data_frame(df_result_stage, BatchSize.UPLOAD_RECORDS.value, init_response.uuid)

        request = API.UploadImageResultsRequest(uuid=init_response.uuid, test_run_id=self.data.id, reset=self._reset)
        finalize_res = krequests.put(
            endpoint_path=API.Path.COMPLETE_UPLOAD_IMAGE_RESULTS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(finalize_res)
        response = from_dict(data_class=API.UploadImageResultsResponse, data=finalize_res.json())
        log.success("uploaded inference results for test run")
        return response.n_uploaded

    @validate_arguments
    def load_remaining_pairs(self, batch_size: int = 10_000_000) -> Tuple[EmbeddingDataFrame, PairDataFrame]:
        """
        Load DataFrames containing computed embeddings and records for each of the image pairs in the configured test
        suite that have not yet had similarity scores computed.

        This method should not be called until all images in the [`TestRun`][kolena.fr.TestRun] have been processed.

        :param batch_size: Optionally specify the maximum number of image pair records to return.
        :return: Two DataFrames, one containing embeddings computed in the previous step (`df_embedding`) and one
            containing records for each of the image pairs that must be computed (`df_pair`). See documentation on
            [`EmbeddingDataFrameSchema`][kolena.fr.datatypes.EmbeddingDataFrameSchema] for expected format when multiple
            embeddings were uploaded from a single image in
            [`TestRun.upload_image_results`][kolena.fr.TestRun.upload_image_results].
        :raises InputValidationError: The requested `batch_size` failed validation.
        :raises RemoteError: Pairs could not be loaded for any reason.
        """
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")

        log.info("loading batch of image pairs for test run")
        init_request = API.InitLoadRemainingPairsRequest(
            test_run_id=self.data.id,
            batch_size=batch_size,
            load_all=self._reset,
        )
        with krequests.put(
            endpoint_path=API.Path.INIT_LOAD_REMAINING_PAIRS.value,
            data=json.dumps(dataclasses.asdict(init_request)),
            stream=True,
        ) as init_res:
            krequests.raise_for_status(init_res)

            load_uuid_embedding = None
            load_uuid_pair = None
            try:
                dfs_embedding = []
                dfs_pair = []
                for line in init_res.iter_lines():
                    partial_response = from_dict(
                        data_class=API.InitLoadRemainingPairsPartialResponse,
                        data=json.loads(line),
                    )
                    load_uuid_embedding = partial_response.embeddings.uuid
                    dfs_embedding.append(_BatchedLoader.load_path(partial_response.embeddings.path, EmbeddingDataFrame))
                    load_uuid_pair = partial_response.pairs.uuid
                    dfs_pair.append(_BatchedLoader.load_path(partial_response.pairs.path, PairDataFrame))

                df_embedding = _BatchedLoader.concat(dfs_embedding, EmbeddingDataFrame)
                df_pair = _BatchedLoader.concat(dfs_pair, PairDataFrame)
                log.info("loaded batch of image pairs for test run")
                return df_embedding, df_pair
            finally:
                for uuid in [load_uuid_embedding, load_uuid_pair]:
                    _BatchedLoader.complete_load(uuid)

    def upload_pair_results(self, df_pair_result: PairResultDataFrame) -> int:
        """
        Upload image pair similarity results for a batch of pairs.

        This method should not be called until all images in the TestRun have been processed.

        All columns except for `image_pair_id` and `similarity` are optional. An empty `similarity` cell in a
        record indicates a pair failure (i.e. one or more of the images in the pair failed to enroll).

        For image pairs containing images with more than one embedding, a single record may be provided with the highest
        similarity score, or `M x N` records may be provided for each embeddings combination in the pair, when there
        are `M` embeddings from `image_a` and `N` embeddings from `image_b`.

        When providing multiple records for a given image pair, use the `embedding_a_index` and `embedding_b_index`
        columns to indicate which embeddings were used to compute a given similarity score. Records for a given image
        pair must be submitted in the same `df_pair_result` DataFrame, and **not** across multiple calls to
        `upload_pair_results`.

        :param df_pair_result: DataFrame containing records describing the similarity score of a pair of images.
        :return: Number of records successfully uploaded.
        :raises TypeValidationError: The DataFrame failed type validation.
        :raises RemoteError: The DataFrame was unable to be successfully ingested for any reason.
        """
        log.info("uploading pair results for test run")
        init_response = init_upload()

        df_validated = validate_df_schema(df_pair_result, PairResultDataFrameSchema)
        validate_df_record_count(df_validated)
        upload_data_frame(df_validated, BatchSize.UPLOAD_RECORDS.value, init_response.uuid)

        request = API.UploadPairResultsRequest(uuid=init_response.uuid, test_run_id=self.data.id, reset=self._reset)
        finalize_res = krequests.put(
            endpoint_path=API.Path.COMPLETE_UPLOAD_PAIR_RESULTS.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(finalize_res)
        log.success("uploaded pair results for test run")
        response = from_dict(data_class=API.UploadPairResultsResponse, data=finalize_res.json())
        return response.n_uploaded


@validate_arguments(config=ValidatorConfig)
def test(model: InferenceModel, test_suite: TestSuite, reset: bool = False) -> None:
    """
    Test the provided [`InferenceModel`][kolena.fr.InferenceModel] on the provided [`TestSuite`][kolena.fr.TestSuite].
    Any tests already in progress for this model on these suites are resumed.

    :param model: The model being tested, implementing both `extract` and `compare` methods.
    :param test_suite: The test suite on which to test the model.
    :param reset: Overwrites existing inferences if set.
    """
    test_run = TestRun(model, test_suite, reset=reset)

    try:
        log.info("starting test run")
        df_image = test_run.load_remaining_images(int(1e12))
        df_image["embedding"] = [
            model.extract(record.locator)
            for record in tqdm(df_image.itertuples(), position=0, leave=True, total=len(df_image))
        ]
        empty_columns = [  # columns not populated in this implementation
            "bounding_box",
            "landmarks_input_image",
            "landmarks",
            "quality_input_image",
            "quality",
            "acceptability",
            "fr_input_image",
            "failure_reason",
        ]
        df_image[empty_columns] = None
        df_image_result = df_image[["image_id", "embedding", *empty_columns]]
        if len(df_image_result) > 0:  # only attempt to upload if this step has not been completed
            test_run.upload_image_results(df_image_result)

        df_embedding, df_pair = test_run.load_remaining_pairs(int(1e12))
        embedding_by_id = {record.image_id: record.embedding for record in df_embedding.itertuples()}
        df_pair["similarity"] = [
            model.compare(embedding_by_id[record.image_a_id], embedding_by_id[record.image_b_id])
            for record in tqdm(df_pair.itertuples(), position=0, leave=True, total=len(df_pair))
        ]
        df_pair_result = df_pair[["image_pair_id", "similarity"]]
        if len(df_pair_result) > 0:
            test_run.upload_pair_results(df_pair_result)
        log.success("completed test run")

    except Exception as e:
        report_crash(test_run.data.id, API.Path.MARK_CRASHED.value)
        raise e



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import pandas as pd
from pydantic import validate_arguments

from kolena._api.v1.core import Model as CoreAPI
from kolena._api.v1.detection import Model as API
from kolena._api.v1.workflow import WorkflowType
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import DFType
from kolena._utils.consts import BatchSize
from kolena._utils.consts import FieldName
from kolena._utils.endpoints import get_model_url
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.detection._internal import BaseTestCase
from kolena.detection._internal import BaseTestImage
from kolena.detection._internal import BaseTestSuite
from kolena.detection._internal.test_image import TestImageType
from kolena.errors import InputValidationError
from kolena.errors import NotFoundError


InferenceType = TypeVar("InferenceType")
SampleInferences = Tuple[TestImageType, Optional[List[InferenceType]]]


class BaseModel(ABC, Frozen, WithTelemetry):
    """
    The base class for [`kolena.classification.Model`][kolena.classification.Model] and
    [`kolena.detection.Model`][kolena.detection.Model].
    """

    name: str
    """
    Unique name of the model, potentially containing information about the architecture, training dataset,
    configuration, framework, commit hash, etc.
    """

    metadata: Dict[str, Any]
    """Unstructured metadata associated with the model."""

    _id: int
    _workflow: WorkflowType

    _TestImageClass: Type[BaseTestImage] = BaseTestImage
    _TestCaseClass: Type[BaseTestCase] = BaseTestCase
    _TestSuiteClass: Type[BaseTestSuite] = BaseTestSuite
    _InferenceClass: Type[InferenceType] = InferenceType
    _LoadInferencesDataFrameClass: Type[DFType] = DFType

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, name: str, workflow: WorkflowType, metadata: Optional[Dict[str, Any]] = None):
        validate_name(name, FieldName.MODEL_NAME)
        try:
            loaded = self._load_by_name(name)
            if len(loaded.metadata) > 0 and loaded.metadata != metadata:
                log.warn(f"mismatch in model metadata, using loaded metadata (loaded: {loaded.metadata})")
        except NotFoundError:
            loaded = self._create(workflow, name, metadata or {})

        self.name = name
        self.metadata = loaded.metadata
        self._id = loaded.id
        self._workflow = WorkflowType(loaded.workflow)
        self._freeze()

    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def _create(cls, workflow: WorkflowType, name: str, metadata: Dict[str, Any]) -> CoreAPI.EntityData:
        request = CoreAPI.CreateRequest(name=name, metadata=metadata, workflow=workflow.value)
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        log.info(f"created model '{name}' ({get_model_url(obj.id)})")
        return obj

    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def _load_by_name(cls, name: str) -> CoreAPI.EntityData:
        request = CoreAPI.LoadByNameRequest(name=name)
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        log.info(f"loaded model '{name}' ({get_model_url(obj.id)})")
        return obj

    @validate_arguments(config=ValidatorConfig)
    def load_inferences(
        self,
        test_object: Union[_TestCaseClass, _TestSuiteClass],
    ) -> List[Tuple[_TestImageClass, Optional[List[_InferenceClass]]]]:
        """
        Retrieve the uploaded inferences with identical ground truth labels for each image in a test case or test suite.

        :return: List of all images and inferences from this model on the provided `test_object`.
        """
        return list(self.iter_inferences(test_object))

    @validate_arguments(config=ValidatorConfig)
    def iter_inferences(
        self,
        test_object: Union[_TestCaseClass, _TestSuiteClass],
    ) -> Iterator[Tuple[_TestImageClass, Optional[List[_InferenceClass]]]]:
        """
        Iterate the uploaded inferences with identical ground truth labels for each image in a test case or test suite.

        :return: Iterator over all images and inferences from this model on the provided `test_object`.
        """
        for df_batch in self._iter_inference_batch_for_reference(test_object):
            yield from (self._inferences_from_record(record) for record in df_batch.itertuples())

    @validate_arguments(config=ValidatorConfig)
    def _iter_inference_batch_for_reference(
        self,
        test_object: Union[_TestCaseClass, _TestSuiteClass],
        batch_size: int = BatchSize.LOAD_SAMPLES.value,
    ) -> Iterator[_LoadInferencesDataFrameClass]:
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")
        test_object_display_type = "test case" if isinstance(test_object, self._TestCaseClass) else "test suite"
        test_object_display_name = f"{test_object_display_type} '{test_object.name}'"
        log.info(f"loading inferences from model '{self.name}' on {test_object_display_name}")
        test_id_key = "test_case_id" if isinstance(test_object, self._TestCaseClass) else "test_suite_id"
        params = dict(model_id=self._id, batch_size=batch_size, **{test_id_key: test_object._id})
        init_request = API.InitLoadInferencesRequest(**params)
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_INFERENCES.value,
            df_class=self._LoadInferencesDataFrameClass,
        )
        log.info(f"loaded inferences from model '{self.name}' on {test_object_display_name}")

    @validate_arguments(config=ValidatorConfig)
    def load_inferences_by_test_case(
        self,
        test_suite: _TestSuiteClass,
    ) -> Dict[int, List[SampleInferences[_TestImageClass, _InferenceClass]]]:
        """Retrieve the uploaded inferences of a test suite for each image, grouped by test case."""
        batches = list(self._iter_inference_batch_for_test_suite(test_suite))
        df_all = pd.concat(batches)
        df = pd.DataFrame(columns=["test_case_id", "test_sample"])
        df["test_case_id"] = df_all["test_case_id"]
        df["test_sample"] = df_all.apply(lambda record: self._inferences_from_record(record), axis=1)
        df_by_testcase = df.groupby("test_case_id")["test_sample"].agg(list).to_dict()
        return df_by_testcase

    @validate_arguments(config=ValidatorConfig)
    def _iter_inference_batch_for_test_suite(
        self,
        test_suite: _TestSuiteClass,
        batch_size: int = BatchSize.LOAD_SAMPLES.value,
    ) -> Iterator[_LoadInferencesDataFrameClass]:
        if batch_size <= 0:
            raise InputValidationError(f"invalid batch_size '{batch_size}': expected positive integer")
        log.info(f"loading inferences from model '{self.name}' on test suite '{test_suite.name}'")
        params = dict(model_id=self._id, batch_size=batch_size, test_suite_id=test_suite._id)
        init_request = API.InitLoadInferencesByTestCaseRequest(**params)
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_INFERENCES_BY_TEST_CASE.value,
            df_class=self._LoadInferencesDataFrameClass,
        )
        log.info(f"loaded inferences from model '{self.name}' on test suite '{test_suite.name}'")

    @abstractmethod
    def _inferences_from_record(self, record: Any) -> Tuple[_TestImageClass, Optional[List[_InferenceClass]]]:
        ...



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np

from kolena._experimental.object_detection import ClassMetricsPerTestCase
from kolena._experimental.object_detection import GroundTruth
from kolena._experimental.object_detection import Inference
from kolena._experimental.object_detection import TestCase
from kolena._experimental.object_detection import TestCaseMetrics
from kolena._experimental.object_detection import TestSample
from kolena._experimental.object_detection import TestSampleMetrics
from kolena._experimental.object_detection import TestSuite
from kolena._experimental.object_detection import TestSuiteMetrics
from kolena._experimental.object_detection import ThresholdConfiguration
from kolena._experimental.object_detection import ThresholdStrategy
from kolena._experimental.object_detection.utils import compute_average_precision
from kolena._experimental.object_detection.utils import compute_confusion_matrix_plot
from kolena._experimental.object_detection.utils import compute_f1_plot_multiclass
from kolena._experimental.object_detection.utils import compute_optimal_f1_threshold_multiclass
from kolena._experimental.object_detection.utils import compute_pr_curve
from kolena._experimental.object_detection.utils import compute_pr_plot_multiclass
from kolena._experimental.object_detection.utils import filter_inferences
from kolena.workflow import Evaluator
from kolena.workflow import Plot
from kolena.workflow.annotation import ScoredClassificationLabel
from kolena.workflow.metrics import f1_score as compute_f1_score
from kolena.workflow.metrics import match_inferences_multiclass
from kolena.workflow.metrics import MulticlassInferenceMatches
from kolena.workflow.metrics import precision as compute_precision
from kolena.workflow.metrics import recall as compute_recall


class MulticlassObjectDetectionEvaluator(Evaluator):
    """
    The `MulticlassObjectDetectionEvaluator` transforms inferences into metrics for the object detection workflow for
    multiple classes.

    When a [`ThresholdConfiguration`][kolena._experimental.object_detection.workflow.ThresholdConfiguration] is
    configured to use an F1-Optimal threshold strategy, the evaluator requires that the first test case retrieved for
    a test suite contains the complete sample set.

    For additional functionality, see the associated [base class documentation][kolena.workflow.evaluator.Evaluator].
    """

    threshold_cache: Dict[str, Dict[str, float]] = {}  # configuration -> label -> threshold
    """
    Assumes that the first test case retrieved for the test suite contains the complete sample set to be used for
    F1-Optimal threshold computation. Subsequent requests for a given threshold strategy (for other test cases) will
    hit this cache and use the previously computed population level confidence thresholds.
    """

    locators_by_test_case: Dict[str, List[str]] = {}
    """
    Keeps track of test sample locators for each test case (used for total # of image count in aggregated metrics).
    """

    matchings_by_test_case: Dict[str, List[MulticlassInferenceMatches]] = defaultdict(list)
    """
    Caches matchings per test case for test case metrics and test case plots.
    """

    def test_sample_metrics_ignored(
        self,
    ) -> TestSampleMetrics:
        return TestSampleMetrics(
            TP_labels=[],
            TP=[],
            FP_labels=[],
            FP=[],
            FN_labels=[],
            FN=[],
            Confused_labels=[],
            Confused=[],
            count_TP=0,
            count_FP=0,
            count_FN=0,
            count_Confused=0,
            has_TP=False,
            has_FP=False,
            has_FN=False,
            has_Confused=False,
            ignored=True,
            max_confidence_above_t=None,
            min_confidence_above_t=None,
            thresholds=[],
            inference_labels=[],
        )

    def test_sample_metrics(
        self,
        bbox_matches: MulticlassInferenceMatches,
        thresholds: Dict[str, float],
    ) -> TestSampleMetrics:
        tp = [inf for _, inf in bbox_matches.matched if inf.score >= thresholds[inf.label]]
        fp = [inf for inf in bbox_matches.unmatched_inf if inf.score >= thresholds[inf.label]]
        fn = [gt for gt, _ in bbox_matches.unmatched_gt] + [
            gt for gt, inf in bbox_matches.matched if inf.score < thresholds[inf.label]
        ]
        confused = [
            inf for _, inf in bbox_matches.unmatched_gt if inf is not None and inf.score >= thresholds[inf.label]
        ]
        non_ignored_inferences = tp + fp
        scores = [inf.score for inf in non_ignored_inferences]
        inference_labels = {inf.label for _, inf in bbox_matches.matched} | {
            inf.label for inf in bbox_matches.unmatched_inf
        }
        fields = [
            ScoredClassificationLabel(label=label, score=thresholds[label])
            for label in sorted(thresholds.keys())
            if label in inference_labels
        ]
        return TestSampleMetrics(
            TP_labels=sorted({inf.label for inf in tp}),
            TP=tp,
            FP_labels=sorted({inf.label for inf in fp}),
            FP=fp,
            FN_labels=sorted({inf.label for inf in fn}),
            FN=fn,
            Confused_labels=sorted({inf.label for inf in confused}),
            Confused=confused,
            count_TP=len(tp),
            count_FP=len(fp),
            count_FN=len(fn),
            count_Confused=len(confused),
            has_TP=len(tp) > 0,
            has_FP=len(fp) > 0,
            has_FN=len(fn) > 0,
            has_Confused=len(confused) > 0,
            ignored=False,
            max_confidence_above_t=max(scores) if len(scores) > 0 else None,
            min_confidence_above_t=min(scores) if len(scores) > 0 else None,
            thresholds=fields,
            inference_labels=sorted(inference_labels),
        )

    def compute_image_metrics(
        self,
        ground_truth: GroundTruth,
        inference: Inference,
        configuration: ThresholdConfiguration,
        test_case_name: str,
    ) -> TestSampleMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        if inference.ignored:
            return self.test_sample_metrics_ignored()

        filtered_inferences = filter_inferences(
            inferences=inference.bboxes,
            confidence_score=configuration.min_confidence_score,
        )
        bbox_matches: MulticlassInferenceMatches = match_inferences_multiclass(
            ground_truth.bboxes,
            filtered_inferences,
            ignored_ground_truths=ground_truth.ignored_bboxes,
            mode="pascal",
            iou_threshold=configuration.iou_threshold,
        )
        self.matchings_by_test_case[test_case_name].append(bbox_matches)

        return self.test_sample_metrics(bbox_matches, thresholds)

    def compute_and_cache_f1_optimal_thresholds(
        self,
        configuration: ThresholdConfiguration,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
    ) -> None:
        if configuration.threshold_strategy != ThresholdStrategy.F1_OPTIMAL:
            return

        if configuration.display_name() in self.threshold_cache.keys():
            return

        all_bbox_matches = [
            match_inferences_multiclass(
                ground_truth.bboxes,
                filter_inferences(inferences=inference.bboxes, confidence_score=configuration.min_confidence_score),
                ignored_ground_truths=ground_truth.ignored_bboxes,
                mode="pascal",
                iou_threshold=configuration.iou_threshold,
            )
            for _, ground_truth, inference in inferences
            if not inference.ignored
        ]
        optimal_thresholds = compute_optimal_f1_threshold_multiclass(all_bbox_matches)
        self.threshold_cache[configuration.display_name()] = defaultdict(
            lambda: configuration.min_confidence_score,
            optimal_thresholds,
        )

    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> List[Tuple[TestSample, TestSampleMetrics]]:
        assert configuration is not None, "must specify configuration"
        # compute thresholds to cache values for subsequent steps
        self.compute_and_cache_f1_optimal_thresholds(configuration, inferences)
        return [(ts, self.compute_image_metrics(gt, inf, configuration, test_case.name)) for ts, gt, inf in inferences]

    def bbox_matches_for_one_label(
        self,
        matchings: List[MulticlassInferenceMatches],
        label: str,
    ) -> Tuple[MulticlassInferenceMatches, int]:
        match_matched = []
        match_unmatched_gt = []
        match_unmatched_inf = []
        samples_count = 0
        # filter the matching to only consider one class
        for match in matchings:
            sample_flag = False
            for gt, inf in match.matched:
                if gt.label == label:
                    sample_flag = True
                    match_matched.append((gt, inf))
            for gt, inf in match.unmatched_gt:
                if gt.label == label:
                    sample_flag = True
                    match_unmatched_gt.append((gt, inf))
            for inf in match.unmatched_inf:
                if inf.label == label:
                    sample_flag = True
                    match_unmatched_inf.append(inf)
            if sample_flag:
                samples_count += 1

        bbox_matches_for_one_label = MulticlassInferenceMatches(
            matched=match_matched,
            unmatched_gt=match_unmatched_gt,
            unmatched_inf=match_unmatched_inf,
        )

        return bbox_matches_for_one_label, samples_count

    def class_metrics_per_test_case(
        self,
        label: str,
        thresholds: Dict[str, float],
        class_matches: MulticlassInferenceMatches,
        samples_count: int,
        average_precision: float,
    ) -> ClassMetricsPerTestCase:
        matched = class_matches.matched
        unmatched_gt = class_matches.unmatched_gt
        unmatched_inf = class_matches.unmatched_inf

        tp = [inf for _, inf in matched if inf.score >= thresholds[inf.label]]
        fp = [inf for inf in unmatched_inf if inf.score >= thresholds[inf.label]]
        fn = [gt for gt, _ in unmatched_gt] + [gt for gt, inf in matched if inf.score < thresholds[inf.label]]
        tp_count = len(tp)
        fp_count = len(fp)
        fn_count = len(fn)
        precision = compute_precision(tp_count, fp_count)
        recall = compute_recall(tp_count, fn_count)
        f1_score = compute_f1_score(tp_count, fp_count, fn_count)

        return ClassMetricsPerTestCase(
            Class=label,
            nImages=samples_count,
            Threshold=thresholds[label],
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            Precision=precision,
            Recall=recall,
            F1=f1_score,
            AP=average_precision,
        )

    def compute_aggregate_label_metrics(
        self,
        matchings: List[MulticlassInferenceMatches],
        label: str,
        thresholds: Dict[str, float],
    ) -> ClassMetricsPerTestCase:
        match_and_count: Tuple[MulticlassInferenceMatches, int] = self.bbox_matches_for_one_label(matchings, label)
        class_matches: MulticlassInferenceMatches = match_and_count[0]
        samples_count: int = match_and_count[1]

        average_precision = 0.0
        baseline_pr_curve = compute_pr_curve([class_matches])
        if baseline_pr_curve is not None:
            average_precision = compute_average_precision(baseline_pr_curve.y, baseline_pr_curve.x)

        return self.class_metrics_per_test_case(label, thresholds, class_matches, samples_count, average_precision)

    def test_case_metrics(
        self,
        per_class_metrics: List[ClassMetricsPerTestCase],
        metrics: List[TestSampleMetrics],
    ) -> TestCaseMetrics:
        tp_count = sum(im.count_TP for im in metrics)
        fp_count = sum(im.count_FP for im in metrics)
        fn_count = sum(im.count_FN for im in metrics)
        ignored_count = sum(1 if im.ignored else 0 for im in metrics)
        return TestCaseMetrics(
            PerClass=per_class_metrics,
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            nIgnored=ignored_count,
            macro_Precision=np.mean([data.Precision for data in per_class_metrics]) if per_class_metrics else 0.0,
            macro_Recall=np.mean([data.Recall for data in per_class_metrics]) if per_class_metrics else 0.0,
            macro_F1=np.mean([data.F1 for data in per_class_metrics]) if per_class_metrics else 0.0,
            mean_AP=np.mean([data.AP for data in per_class_metrics]) if per_class_metrics else 0.0,
        )

    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[TestSampleMetrics],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> TestCaseMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        all_bbox_matches = self.matchings_by_test_case[test_case.name]
        self.locators_by_test_case[test_case.name] = [ts.locator for ts, _, _ in inferences]

        # compute nested metrics per class
        labels = {gt.label for _, gts, _ in inferences for gt in gts.bboxes}
        per_class_metrics: List[ClassMetricsPerTestCase] = []
        for label in sorted(labels):
            metrics_per_class = self.compute_aggregate_label_metrics(all_bbox_matches, label, thresholds)
            per_class_metrics.append(metrics_per_class)

        return self.test_case_metrics(per_class_metrics, metrics)

    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[TestSampleMetrics],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> Optional[List[Plot]]:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        all_bbox_matches = self.matchings_by_test_case[test_case.name]

        # clean matching for confusion matrix
        match_matched = [
            (gt, inf) for match in all_bbox_matches for gt, inf in match.matched if inf.score >= thresholds[inf.label]
        ]
        match_unmatched_gt = [
            (gt, inf)
            for match in all_bbox_matches
            for gt, inf in match.unmatched_gt
            if inf is not None and inf.score >= thresholds[inf.label]
        ]
        confusion_matrix_matchings = [
            MulticlassInferenceMatches(
                matched=match_matched,
                unmatched_gt=match_unmatched_gt,
                unmatched_inf=[],
            ),
        ]

        plots: Optional[List[Plot]] = []
        plots.extend(
            filter(
                None,
                [
                    compute_f1_plot_multiclass(all_bbox_matches),
                    compute_pr_plot_multiclass(all_bbox_matches),
                    compute_confusion_matrix_plot(confusion_matrix_matchings),
                ],
            ),
        )

        return plots

    def test_suite_metrics(self, unique_locators: Set[str], average_precisions: List[float]) -> TestSuiteMetrics:
        return TestSuiteMetrics(
            n_images=len(unique_locators),
            mean_AP=np.mean(average_precisions) if average_precisions else 0.0,
        )

    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, TestCaseMetrics]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> TestSuiteMetrics:
        assert configuration is not None, "must specify configuration"
        unique_locators = {locator for tc, _ in metrics for locator in self.locators_by_test_case[tc.name]}
        average_precisions = [tcm.mean_AP for _, tcm in metrics]
        return self.test_suite_metrics(unique_locators, average_precisions)

    def get_confidence_thresholds(self, configuration: ThresholdConfiguration) -> Dict[str, float]:
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_03:
            return defaultdict(lambda: 0.3)
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_05:
            return defaultdict(lambda: 0.5)
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_075:
            return defaultdict(lambda: 0.75)
        if configuration.threshold_strategy == ThresholdStrategy.F1_OPTIMAL:
            return self.threshold_cache[configuration.display_name()]
        raise RuntimeError(f"unrecognized threshold strategy: {configuration.threshold_strategy}")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional

from pydantic import validate_arguments

from kolena._utils.validators import ValidatorConfig
from kolena.classification.multiclass import Model
from kolena.classification.multiclass import MulticlassClassificationEvaluator
from kolena.classification.multiclass import TestSuite
from kolena.classification.multiclass.workflow import ThresholdConfiguration
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow.test_run import test as base_test
from kolena.workflow.test_run import TestRun as BaseTestRun


class TestRun(BaseTestRun):
    """
    Convenience alias for [`kolena.workflow.TestRun`][kolena.workflow.test_run.TestRun], configured for the
    `kolena.classification.multiclass` workflow and evaluator.

    Interface to run tests for a [`Model`][kolena.classification.multiclass.Model] on a
    [`TestSuite`][kolena.classification.multiclass.TestSuite].

    For a streamlined interface, see [`test`][kolena.classification.multiclass.test].

    :param model: The model being tested.
    :param test_suite: The test suite on which to test the model.
    :param reset: Overwrites existing inferences if set.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        model: Model,
        test_suite: TestSuite,
        configurations: Optional[List[EvaluatorConfiguration]] = None,
        reset: bool = False,
    ):
        if configurations is None:
            configurations = [ThresholdConfiguration()]
        super().__init__(
            model=model,
            test_suite=test_suite,
            evaluator=MulticlassClassificationEvaluator,
            configurations=configurations,
            reset=reset,
        )


@validate_arguments(config=ValidatorConfig)
def test(
    model: Model,
    test_suite: TestSuite,
    configurations: Optional[List[EvaluatorConfiguration]] = None,
    reset: bool = False,
) -> None:
    """
    Convenience alias for [`test`][kolena.workflow.test] configured for the `kolena.classification.multiclass` workflow
    and evaluator.

    Tests the provided [`Model`][kolena.classification.multiclass.Model] on the provided
    [`TestSuite`][kolena.classification.multiclass.TestSuite]. Any test already in progress for this model on this
    test suite is resumed.

    :param model: The model being tested, implementing the `infer` method.
    :param test_suite: The test suite on which to test the model.
    :param configurations: An optional list of configurations to use when running the evaluator. Defaults to selecting
        the max confidence label (with no thresholding) if unset.
    :param reset: Overwrites existing inferences if set.
    """
    if configurations is None:
        configurations = [ThresholdConfiguration()]
    base_test(
        model=model,
        test_suite=test_suite,
        evaluator=MulticlassClassificationEvaluator,
        configurations=configurations,
        reset=reset,
    )
